import itertools

SPDX_LICENSES = {
    "0BSD": "https://spdx.org/licenses/0BSD.html",
    "AAL": "https://spdx.org/licenses/AAL.html",
    "ADSL": "https://spdx.org/licenses/ADSL.html",
    "AFL-1.1": "https://spdx.org/licenses/AFL-1.1.html",
    "AFL-1.2": "https://spdx.org/licenses/AFL-1.2.html",
    "AFL-2.0": "https://spdx.org/licenses/AFL-2.0.html",
    "AFL-2.1": "https://spdx.org/licenses/AFL-2.1.html",
    "AFL-3.0": "https://spdx.org/licenses/AFL-3.0.html",
    "AGPL-1.0": "https://spdx.org/licenses/AGPL-1.0.html",
    "AGPL-1.0-only": "https://spdx.org/licenses/AGPL-1.0-only.html",
    "AGPL-1.0-or-later": "https://spdx.org/licenses/AGPL-1.0-or-later.html",
    "AGPL-3.0": "https://spdx.org/licenses/AGPL-3.0.html",
    "AGPL-3.0-only": "https://spdx.org/licenses/AGPL-3.0-only.html",
    "AGPL-3.0-or-later": "https://spdx.org/licenses/AGPL-3.0-or-later.html",
    "AMDPLPA": "https://spdx.org/licenses/AMDPLPA.html",
    "AML": "https://spdx.org/licenses/AML.html",
    "AMPAS": "https://spdx.org/licenses/AMPAS.html",
    "ANTLR-PD": "https://spdx.org/licenses/ANTLR-PD.html",
    "APAFML": "https://spdx.org/licenses/APAFML.html",
    "APL-1.0": "https://spdx.org/licenses/APL-1.0.html",
    "APSL-1.0": "https://spdx.org/licenses/APSL-1.0.html",
    "APSL-1.1": "https://spdx.org/licenses/APSL-1.1.html",
    "APSL-1.2": "https://spdx.org/licenses/APSL-1.2.html",
    "APSL-2.0": "https://spdx.org/licenses/APSL-2.0.html",
    "Abstyles": "https://spdx.org/licenses/Abstyles.html",
    "Adobe-2006": "https://spdx.org/licenses/Adobe-2006.html",
    "Adobe-Glyph": "https://spdx.org/licenses/Adobe-Glyph.html",
    "Afmparse": "https://spdx.org/licenses/Afmparse.html",
    "Aladdin": "https://spdx.org/licenses/Aladdin.html",
    "Apache-1.0": "https://spdx.org/licenses/Apache-1.0.html",
    "Apache-1.1": "https://spdx.org/licenses/Apache-1.1.html",
    "Apache-2.0": "https://spdx.org/licenses/Apache-2.0.html",
    "Artistic-1.0": "https://spdx.org/licenses/Artistic-1.0.html",
    "Artistic-1.0-Perl": "https://spdx.org/licenses/Artistic-1.0-Perl.html",
    "Artistic-1.0-cl8": "https://spdx.org/licenses/Artistic-1.0-cl8.html",
    "Artistic-2.0": "https://spdx.org/licenses/Artistic-2.0.html",
    "BSD-1-Clause": "https://spdx.org/licenses/BSD-1-Clause.html",
    "BSD-2-Clause": "https://spdx.org/licenses/BSD-2-Clause.html",
    "BSD-2-Clause-FreeBSD": "https://spdx.org/licenses/BSD-2-Clause-FreeBSD.html",
    "BSD-2-Clause-NetBSD": "https://spdx.org/licenses/BSD-2-Clause-NetBSD.html",
    "BSD-2-Clause-Patent": "https://spdx.org/licenses/BSD-2-Clause-Patent.html",
    "BSD-3-Clause": "https://spdx.org/licenses/BSD-3-Clause.html",
    "BSD-3-Clause-Attribution": "https://spdx.org/licenses/BSD-3-Clause-Attribution.html",
    "BSD-3-Clause-Clear": "https://spdx.org/licenses/BSD-3-Clause-Clear.html",
    "BSD-3-Clause-LBNL": "https://spdx.org/licenses/BSD-3-Clause-LBNL.html",
    "BSD-3-Clause-No-Nuclear-License": "https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License.html",
    "BSD-3-Clause-No-Nuclear-License-2014": "https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License-2014.html",
    "BSD-3-Clause-No-Nuclear-Warranty": "https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-Warranty.html",
    "BSD-4-Clause": "https://spdx.org/licenses/BSD-4-Clause.html",
    "BSD-4-Clause-UC": "https://spdx.org/licenses/BSD-4-Clause-UC.html",
    "BSD-Protection": "https://spdx.org/licenses/BSD-Protection.html",
    "BSD-Source-Code": "https://spdx.org/licenses/BSD-Source-Code.html",
    "BSL-1.0": "https://spdx.org/licenses/BSL-1.0.html",
    "Bahyph": "https://spdx.org/licenses/Bahyph.html",
    "Barr": "https://spdx.org/licenses/Barr.html",
    "Beerware": "https://spdx.org/licenses/Beerware.html",
    "BitTorrent-1.0": "https://spdx.org/licenses/BitTorrent-1.0.html",
    "BitTorrent-1.1": "https://spdx.org/licenses/BitTorrent-1.1.html",
    "BlueOak-1.0.0": "https://spdx.org/licenses/BlueOak-1.0.0.html",
    "Borceux": "https://spdx.org/licenses/Borceux.html",
    "CATOSL-1.1": "https://spdx.org/licenses/CATOSL-1.1.html",
    "CC-BY-1.0": "https://spdx.org/licenses/CC-BY-1.0.html",
    "CC-BY-2.0": "https://spdx.org/licenses/CC-BY-2.0.html",
    "CC-BY-2.5": "https://spdx.org/licenses/CC-BY-2.5.html",
    "CC-BY-3.0": "https://spdx.org/licenses/CC-BY-3.0.html",
    "CC-BY-4.0": "https://spdx.org/licenses/CC-BY-4.0.html",
    "CC-BY-NC-1.0": "https://spdx.org/licenses/CC-BY-NC-1.0.html",
    "CC-BY-NC-2.0": "https://spdx.org/licenses/CC-BY-NC-2.0.html",
    "CC-BY-NC-2.5": "https://spdx.org/licenses/CC-BY-NC-2.5.html",
    "CC-BY-NC-3.0": "https://spdx.org/licenses/CC-BY-NC-3.0.html",
    "CC-BY-NC-4.0": "https://spdx.org/licenses/CC-BY-NC-4.0.html",
    "CC-BY-NC-ND-1.0": "https://spdx.org/licenses/CC-BY-NC-ND-1.0.html",
    "CC-BY-NC-ND-2.0": "https://spdx.org/licenses/CC-BY-NC-ND-2.0.html",
    "CC-BY-NC-ND-2.5": "https://spdx.org/licenses/CC-BY-NC-ND-2.5.html",
    "CC-BY-NC-ND-3.0": "https://spdx.org/licenses/CC-BY-NC-ND-3.0.html",
    "CC-BY-NC-ND-4.0": "https://spdx.org/licenses/CC-BY-NC-ND-4.0.html",
    "CC-BY-NC-SA-1.0": "https://spdx.org/licenses/CC-BY-NC-SA-1.0.html",
    "CC-BY-NC-SA-2.0": "https://spdx.org/licenses/CC-BY-NC-SA-2.0.html",
    "CC-BY-NC-SA-2.5": "https://spdx.org/licenses/CC-BY-NC-SA-2.5.html",
    "CC-BY-NC-SA-3.0": "https://spdx.org/licenses/CC-BY-NC-SA-3.0.html",
    "CC-BY-NC-SA-4.0": "https://spdx.org/licenses/CC-BY-NC-SA-4.0.html",
    "CC-BY-ND-1.0": "https://spdx.org/licenses/CC-BY-ND-1.0.html",
    "CC-BY-ND-2.0": "https://spdx.org/licenses/CC-BY-ND-2.0.html",
    "CC-BY-ND-2.5": "https://spdx.org/licenses/CC-BY-ND-2.5.html",
    "CC-BY-ND-3.0": "https://spdx.org/licenses/CC-BY-ND-3.0.html",
    "CC-BY-ND-4.0": "https://spdx.org/licenses/CC-BY-ND-4.0.html",
    "CC-BY-SA-1.0": "https://spdx.org/licenses/CC-BY-SA-1.0.html",
    "CC-BY-SA-2.0": "https://spdx.org/licenses/CC-BY-SA-2.0.html",
    "CC-BY-SA-2.5": "https://spdx.org/licenses/CC-BY-SA-2.5.html",
    "CC-BY-SA-3.0": "https://spdx.org/licenses/CC-BY-SA-3.0.html",
    "CC-BY-SA-4.0": "https://spdx.org/licenses/CC-BY-SA-4.0.html",
    "CC0-1.0": "https://spdx.org/licenses/CC0-1.0.html",
    "CDDL-1.0": "https://spdx.org/licenses/CDDL-1.0.html",
    "CDDL-1.1": "https://spdx.org/licenses/CDDL-1.1.html",
    "CDLA-Permissive-1.0": "https://spdx.org/licenses/CDLA-Permissive-1.0.html",
    "CDLA-Sharing-1.0": "https://spdx.org/licenses/CDLA-Sharing-1.0.html",
    "CECILL-1.0": "https://spdx.org/licenses/CECILL-1.0.html",
    "CECILL-1.1": "https://spdx.org/licenses/CECILL-1.1.html",
    "CECILL-2.0": "https://spdx.org/licenses/CECILL-2.0.html",
    "CECILL-2.1": "https://spdx.org/licenses/CECILL-2.1.html",
    "CECILL-B": "https://spdx.org/licenses/CECILL-B.html",
    "CECILL-C": "https://spdx.org/licenses/CECILL-C.html",
    "CERN-OHL-1.1": "https://spdx.org/licenses/CERN-OHL-1.1.html",
    "CERN-OHL-1.2": "https://spdx.org/licenses/CERN-OHL-1.2.html",
    "CNRI-Jython": "https://spdx.org/licenses/CNRI-Jython.html",
    "CNRI-Python": "https://spdx.org/licenses/CNRI-Python.html",
    "CNRI-Python-GPL-Compatible": "https://spdx.org/licenses/CNRI-Python-GPL-Compatible.html",
    "CPAL-1.0": "https://spdx.org/licenses/CPAL-1.0.html",
    "CPL-1.0": "https://spdx.org/licenses/CPL-1.0.html",
    "CPOL-1.02": "https://spdx.org/licenses/CPOL-1.02.html",
    "CUA-OPL-1.0": "https://spdx.org/licenses/CUA-OPL-1.0.html",
    "Caldera": "https://spdx.org/licenses/Caldera.html",
    "ClArtistic": "https://spdx.org/licenses/ClArtistic.html",
    "Condor-1.1": "https://spdx.org/licenses/Condor-1.1.html",
    "Crossword": "https://spdx.org/licenses/Crossword.html",
    "CrystalStacker": "https://spdx.org/licenses/CrystalStacker.html",
    "Cube": "https://spdx.org/licenses/Cube.html",
    "D-FSL-1.0": "https://spdx.org/licenses/D-FSL-1.0.html",
    "DOC": "https://spdx.org/licenses/DOC.html",
    "DSDP": "https://spdx.org/licenses/DSDP.html",
    "Dotseqn": "https://spdx.org/licenses/Dotseqn.html",
    "ECL-1.0": "https://spdx.org/licenses/ECL-1.0.html",
    "ECL-2.0": "https://spdx.org/licenses/ECL-2.0.html",
    "EFL-1.0": "https://spdx.org/licenses/EFL-1.0.html",
    "EFL-2.0": "https://spdx.org/licenses/EFL-2.0.html",
    "EPL-1.0": "https://spdx.org/licenses/EPL-1.0.html",
    "EPL-2.0": "https://spdx.org/licenses/EPL-2.0.html",
    "EUDatagrid": "https://spdx.org/licenses/EUDatagrid.html",
    "EUPL-1.0": "https://spdx.org/licenses/EUPL-1.0.html",
    "EUPL-1.1": "https://spdx.org/licenses/EUPL-1.1.html",
    "EUPL-1.2": "https://spdx.org/licenses/EUPL-1.2.html",
    "Entessa": "https://spdx.org/licenses/Entessa.html",
    "ErlPL-1.1": "https://spdx.org/licenses/ErlPL-1.1.html",
    "Eurosym": "https://spdx.org/licenses/Eurosym.html",
    "FSFAP": "https://spdx.org/licenses/FSFAP.html",
    "FSFUL": "https://spdx.org/licenses/FSFUL.html",
    "FSFULLR": "https://spdx.org/licenses/FSFULLR.html",
    "FTL": "https://spdx.org/licenses/FTL.html",
    "Fair": "https://spdx.org/licenses/Fair.html",
    "Frameworx-1.0": "https://spdx.org/licenses/Frameworx-1.0.html",
    "FreeImage": "https://spdx.org/licenses/FreeImage.html",
    "GFDL-1.1": "https://spdx.org/licenses/GFDL-1.1.html",
    "GFDL-1.1-only": "https://spdx.org/licenses/GFDL-1.1-only.html",
    "GFDL-1.1-or-later": "https://spdx.org/licenses/GFDL-1.1-or-later.html",
    "GFDL-1.2": "https://spdx.org/licenses/GFDL-1.2.html",
    "GFDL-1.2-only": "https://spdx.org/licenses/GFDL-1.2-only.html",
    "GFDL-1.2-or-later": "https://spdx.org/licenses/GFDL-1.2-or-later.html",
    "GFDL-1.3": "https://spdx.org/licenses/GFDL-1.3.html",
    "GFDL-1.3-only": "https://spdx.org/licenses/GFDL-1.3-only.html",
    "GFDL-1.3-or-later": "https://spdx.org/licenses/GFDL-1.3-or-later.html",
    "GL2PS": "https://spdx.org/licenses/GL2PS.html",
    "GPL-1.0": "https://spdx.org/licenses/GPL-1.0.html",
    "GPL-1.0+": "https://spdx.org/licenses/GPL-1.0+.html",
    "GPL-1.0-only": "https://spdx.org/licenses/GPL-1.0-only.html",
    "GPL-1.0-or-later": "https://spdx.org/licenses/GPL-1.0-or-later.html",
    "GPL-2.0": "https://spdx.org/licenses/GPL-2.0.html",
    "GPL-2.0+": "https://spdx.org/licenses/GPL-2.0+.html",
    "GPL-2.0-only": "https://spdx.org/licenses/GPL-2.0-only.html",
    "GPL-2.0-or-later": "https://spdx.org/licenses/GPL-2.0-or-later.html",
    "GPL-2.0-with-GCC-exception": "https://spdx.org/licenses/GPL-2.0-with-GCC-exception.html",
    "GPL-2.0-with-autoconf-exception": "https://spdx.org/licenses/GPL-2.0-with-autoconf-exception.html",
    "GPL-2.0-with-bison-exception": "https://spdx.org/licenses/GPL-2.0-with-bison-exception.html",
    "GPL-2.0-with-classpath-exception": "https://spdx.org/licenses/GPL-2.0-with-classpath-exception.html",
    "GPL-2.0-with-font-exception": "https://spdx.org/licenses/GPL-2.0-with-font-exception.html",
    "GPL-3.0": "https://spdx.org/licenses/GPL-3.0.html",
    "GPL-3.0+": "https://spdx.org/licenses/GPL-3.0+.html",
    "GPL-3.0-only": "https://spdx.org/licenses/GPL-3.0-only.html",
    "GPL-3.0-or-later": "https://spdx.org/licenses/GPL-3.0-or-later.html",
    "GPL-3.0-with-GCC-exception": "https://spdx.org/licenses/GPL-3.0-with-GCC-exception.html",
    "GPL-3.0-with-autoconf-exception": "https://spdx.org/licenses/GPL-3.0-with-autoconf-exception.html",
    "Giftware": "https://spdx.org/licenses/Giftware.html",
    "Glide": "https://spdx.org/licenses/Glide.html",
    "Glulxe": "https://spdx.org/licenses/Glulxe.html",
    "HPND": "https://spdx.org/licenses/HPND.html",
    "HPND-sell-variant": "https://spdx.org/licenses/HPND-sell-variant.html",
    "HaskellReport": "https://spdx.org/licenses/HaskellReport.html",
    "IBM-pibs": "https://spdx.org/licenses/IBM-pibs.html",
    "ICU": "https://spdx.org/licenses/ICU.html",
    "IJG": "https://spdx.org/licenses/IJG.html",
    "IPA": "https://spdx.org/licenses/IPA.html",
    "IPL-1.0": "https://spdx.org/licenses/IPL-1.0.html",
    "ISC": "https://spdx.org/licenses/ISC.html",
    "ImageMagick": "https://spdx.org/licenses/ImageMagick.html",
    "Imlib2": "https://spdx.org/licenses/Imlib2.html",
    "Info-ZIP": "https://spdx.org/licenses/Info-ZIP.html",
    "Intel": "https://spdx.org/licenses/Intel.html",
    "Intel-ACPI": "https://spdx.org/licenses/Intel-ACPI.html",
    "Interbase-1.0": "https://spdx.org/licenses/Interbase-1.0.html",
    "JPNIC": "https://spdx.org/licenses/JPNIC.html",
    "JSON": "https://spdx.org/licenses/JSON.html",
    "JasPer-2.0": "https://spdx.org/licenses/JasPer-2.0.html",
    "LAL-1.2": "https://spdx.org/licenses/LAL-1.2.html",
    "LAL-1.3": "https://spdx.org/licenses/LAL-1.3.html",
    "LGPL-2.0": "https://spdx.org/licenses/LGPL-2.0.html",
    "LGPL-2.0+": "https://spdx.org/licenses/LGPL-2.0+.html",
    "LGPL-2.0-only": "https://spdx.org/licenses/LGPL-2.0-only.html",
    "LGPL-2.0-or-later": "https://spdx.org/licenses/LGPL-2.0-or-later.html",
    "LGPL-2.1": "https://spdx.org/licenses/LGPL-2.1.html",
    "LGPL-2.1+": "https://spdx.org/licenses/LGPL-2.1+.html",
    "LGPL-2.1-only": "https://spdx.org/licenses/LGPL-2.1-only.html",
    "LGPL-2.1-or-later": "https://spdx.org/licenses/LGPL-2.1-or-later.html",
    "LGPL-3.0": "https://spdx.org/licenses/LGPL-3.0.html",
    "LGPL-3.0+": "https://spdx.org/licenses/LGPL-3.0+.html",
    "LGPL-3.0-only": "https://spdx.org/licenses/LGPL-3.0-only.html",
    "LGPL-3.0-or-later": "https://spdx.org/licenses/LGPL-3.0-or-later.html",
    "LGPLLR": "https://spdx.org/licenses/LGPLLR.html",
    "LPL-1.0": "https://spdx.org/licenses/LPL-1.0.html",
    "LPL-1.02": "https://spdx.org/licenses/LPL-1.02.html",
    "LPPL-1.0": "https://spdx.org/licenses/LPPL-1.0.html",
    "LPPL-1.1": "https://spdx.org/licenses/LPPL-1.1.html",
    "LPPL-1.2": "https://spdx.org/licenses/LPPL-1.2.html",
    "LPPL-1.3a": "https://spdx.org/licenses/LPPL-1.3a.html",
    "LPPL-1.3c": "https://spdx.org/licenses/LPPL-1.3c.html",
    "Latex2e": "https://spdx.org/licenses/Latex2e.html",
    "Leptonica": "https://spdx.org/licenses/Leptonica.html",
    "LiLiQ-P-1.1": "https://spdx.org/licenses/LiLiQ-P-1.1.html",
    "LiLiQ-R-1.1": "https://spdx.org/licenses/LiLiQ-R-1.1.html",
    "LiLiQ-Rplus-1.1": "https://spdx.org/licenses/LiLiQ-Rplus-1.1.html",
    "Libpng": "https://spdx.org/licenses/Libpng.html",
    "Linux-OpenIB": "https://spdx.org/licenses/Linux-OpenIB.html",
    "MIT": "https://spdx.org/licenses/MIT.html",
    "MIT-0": "https://spdx.org/licenses/MIT-0.html",
    "MIT-CMU": "https://spdx.org/licenses/MIT-CMU.html",
    "MIT-advertising": "https://spdx.org/licenses/MIT-advertising.html",
    "MIT-enna": "https://spdx.org/licenses/MIT-enna.html",
    "MIT-feh": "https://spdx.org/licenses/MIT-feh.html",
    "MITNFA": "https://spdx.org/licenses/MITNFA.html",
    "MPL-1.0": "https://spdx.org/licenses/MPL-1.0.html",
    "MPL-1.1": "https://spdx.org/licenses/MPL-1.1.html",
    "MPL-2.0": "https://spdx.org/licenses/MPL-2.0.html",
    "MPL-2.0-no-copyleft-exception": "https://spdx.org/licenses/MPL-2.0-no-copyleft-exception.html",
    "MS-PL": "https://spdx.org/licenses/MS-PL.html",
    "MS-RL": "https://spdx.org/licenses/MS-RL.html",
    "MTLL": "https://spdx.org/licenses/MTLL.html",
    "MakeIndex": "https://spdx.org/licenses/MakeIndex.html",
    "MirOS": "https://spdx.org/licenses/MirOS.html",
    "Motosoto": "https://spdx.org/licenses/Motosoto.html",
    "Multics": "https://spdx.org/licenses/Multics.html",
    "Mup": "https://spdx.org/licenses/Mup.html",
    "NASA-1.3": "https://spdx.org/licenses/NASA-1.3.html",
    "NBPL-1.0": "https://spdx.org/licenses/NBPL-1.0.html",
    "NCSA": "https://spdx.org/licenses/NCSA.html",
    "NGPL": "https://spdx.org/licenses/NGPL.html",
    "NLOD-1.0": "https://spdx.org/licenses/NLOD-1.0.html",
    "NLPL": "https://spdx.org/licenses/NLPL.html",
    "NOSL": "https://spdx.org/licenses/NOSL.html",
    "NPL-1.0": "https://spdx.org/licenses/NPL-1.0.html",
    "NPL-1.1": "https://spdx.org/licenses/NPL-1.1.html",
    "NPOSL-3.0": "https://spdx.org/licenses/NPOSL-3.0.html",
    "NRL": "https://spdx.org/licenses/NRL.html",
    "NTP": "https://spdx.org/licenses/NTP.html",
    "Naumen": "https://spdx.org/licenses/Naumen.html",
    "Net-SNMP": "https://spdx.org/licenses/Net-SNMP.html",
    "NetCDF": "https://spdx.org/licenses/NetCDF.html",
    "Newsletr": "https://spdx.org/licenses/Newsletr.html",
    "Nokia": "https://spdx.org/licenses/Nokia.html",
    "Noweb": "https://spdx.org/licenses/Noweb.html",
    "Nunit": "https://spdx.org/licenses/Nunit.html",
    "OCCT-PL": "https://spdx.org/licenses/OCCT-PL.html",
    "OCLC-2.0": "https://spdx.org/licenses/OCLC-2.0.html",
    "ODC-By-1.0": "https://spdx.org/licenses/ODC-By-1.0.html",
    "ODbL-1.0": "https://spdx.org/licenses/ODbL-1.0.html",
    "OFL-1.0": "https://spdx.org/licenses/OFL-1.0.html",
    "OFL-1.1": "https://spdx.org/licenses/OFL-1.1.html",
    "OGL-UK-1.0": "https://spdx.org/licenses/OGL-UK-1.0.html",
    "OGL-UK-2.0": "https://spdx.org/licenses/OGL-UK-2.0.html",
    "OGL-UK-3.0": "https://spdx.org/licenses/OGL-UK-3.0.html",
    "OGTSL": "https://spdx.org/licenses/OGTSL.html",
    "OLDAP-1.1": "https://spdx.org/licenses/OLDAP-1.1.html",
    "OLDAP-1.2": "https://spdx.org/licenses/OLDAP-1.2.html",
    "OLDAP-1.3": "https://spdx.org/licenses/OLDAP-1.3.html",
    "OLDAP-1.4": "https://spdx.org/licenses/OLDAP-1.4.html",
    "OLDAP-2.0": "https://spdx.org/licenses/OLDAP-2.0.html",
    "OLDAP-2.0.1": "https://spdx.org/licenses/OLDAP-2.0.1.html",
    "OLDAP-2.1": "https://spdx.org/licenses/OLDAP-2.1.html",
    "OLDAP-2.2": "https://spdx.org/licenses/OLDAP-2.2.html",
    "OLDAP-2.2.1": "https://spdx.org/licenses/OLDAP-2.2.1.html",
    "OLDAP-2.2.2": "https://spdx.org/licenses/OLDAP-2.2.2.html",
    "OLDAP-2.3": "https://spdx.org/licenses/OLDAP-2.3.html",
    "OLDAP-2.4": "https://spdx.org/licenses/OLDAP-2.4.html",
    "OLDAP-2.5": "https://spdx.org/licenses/OLDAP-2.5.html",
    "OLDAP-2.6": "https://spdx.org/licenses/OLDAP-2.6.html",
    "OLDAP-2.7": "https://spdx.org/licenses/OLDAP-2.7.html",
    "OLDAP-2.8": "https://spdx.org/licenses/OLDAP-2.8.html",
    "OML": "https://spdx.org/licenses/OML.html",
    "OPL-1.0": "https://spdx.org/licenses/OPL-1.0.html",
    "OSET-PL-2.1": "https://spdx.org/licenses/OSET-PL-2.1.html",
    "OSL-1.0": "https://spdx.org/licenses/OSL-1.0.html",
    "OSL-1.1": "https://spdx.org/licenses/OSL-1.1.html",
    "OSL-2.0": "https://spdx.org/licenses/OSL-2.0.html",
    "OSL-2.1": "https://spdx.org/licenses/OSL-2.1.html",
    "OSL-3.0": "https://spdx.org/licenses/OSL-3.0.html",
    "OpenSSL": "https://spdx.org/licenses/OpenSSL.html",
    "PDDL-1.0": "https://spdx.org/licenses/PDDL-1.0.html",
    "PHP-3.0": "https://spdx.org/licenses/PHP-3.0.html",
    "PHP-3.01": "https://spdx.org/licenses/PHP-3.01.html",
    "Plexus": "https://spdx.org/licenses/Plexus.html",
    "PostgreSQL": "https://spdx.org/licenses/PostgreSQL.html",
    "Python-2.0": "https://spdx.org/licenses/Python-2.0.html",
    "QPL-1.0": "https://spdx.org/licenses/QPL-1.0.html",
    "Qhull": "https://spdx.org/licenses/Qhull.html",
    "RHeCos-1.1": "https://spdx.org/licenses/RHeCos-1.1.html",
    "RPL-1.1": "https://spdx.org/licenses/RPL-1.1.html",
    "RPL-1.5": "https://spdx.org/licenses/RPL-1.5.html",
    "RPSL-1.0": "https://spdx.org/licenses/RPSL-1.0.html",
    "RSA-MD": "https://spdx.org/licenses/RSA-MD.html",
    "RSCPL": "https://spdx.org/licenses/RSCPL.html",
    "Rdisc": "https://spdx.org/licenses/Rdisc.html",
    "Ruby": "https://spdx.org/licenses/Ruby.html",
    "SAX-PD": "https://spdx.org/licenses/SAX-PD.html",
    "SCEA": "https://spdx.org/licenses/SCEA.html",
    "SGI-B-1.0": "https://spdx.org/licenses/SGI-B-1.0.html",
    "SGI-B-1.1": "https://spdx.org/licenses/SGI-B-1.1.html",
    "SGI-B-2.0": "https://spdx.org/licenses/SGI-B-2.0.html",
    "SISSL": "https://spdx.org/licenses/SISSL.html",
    "SISSL-1.2": "https://spdx.org/licenses/SISSL-1.2.html",
    "SMLNJ": "https://spdx.org/licenses/SMLNJ.html",
    "SMPPL": "https://spdx.org/licenses/SMPPL.html",
    "SNIA": "https://spdx.org/licenses/SNIA.html",
    "SPL-1.0": "https://spdx.org/licenses/SPL-1.0.html",
    "SWL": "https://spdx.org/licenses/SWL.html",
    "Saxpath": "https://spdx.org/licenses/Saxpath.html",
    "Sendmail": "https://spdx.org/licenses/Sendmail.html",
    "Sendmail-8.23": "https://spdx.org/licenses/Sendmail-8.23.html",
    "SimPL-2.0": "https://spdx.org/licenses/SimPL-2.0.html",
    "Sleepycat": "https://spdx.org/licenses/Sleepycat.html",
    "Spencer-86": "https://spdx.org/licenses/Spencer-86.html",
    "Spencer-94": "https://spdx.org/licenses/Spencer-94.html",
    "Spencer-99": "https://spdx.org/licenses/Spencer-99.html",
    "StandardML-NJ": "https://spdx.org/licenses/StandardML-NJ.html",
    "SugarCRM-1.1.3": "https://spdx.org/licenses/SugarCRM-1.1.3.html",
    "TAPR-OHL-1.0": "https://spdx.org/licenses/TAPR-OHL-1.0.html",
    "TCL": "https://spdx.org/licenses/TCL.html",
    "TCP-wrappers": "https://spdx.org/licenses/TCP-wrappers.html",
    "TMate": "https://spdx.org/licenses/TMate.html",
    "TORQUE-1.1": "https://spdx.org/licenses/TORQUE-1.1.html",
    "TOSL": "https://spdx.org/licenses/TOSL.html",
    "TU-Berlin-1.0": "https://spdx.org/licenses/TU-Berlin-1.0.html",
    "TU-Berlin-2.0": "https://spdx.org/licenses/TU-Berlin-2.0.html",
    "UPL-1.0": "https://spdx.org/licenses/UPL-1.0.html",
    "Unicode-DFS-2015": "https://spdx.org/licenses/Unicode-DFS-2015.html",
    "Unicode-DFS-2016": "https://spdx.org/licenses/Unicode-DFS-2016.html",
    "Unicode-TOU": "https://spdx.org/licenses/Unicode-TOU.html",
    "Unlicense": "https://spdx.org/licenses/Unlicense.html",
    "VOSTROM": "https://spdx.org/licenses/VOSTROM.html",
    "VSL-1.0": "https://spdx.org/licenses/VSL-1.0.html",
    "Vim": "https://spdx.org/licenses/Vim.html",
    "W3C": "https://spdx.org/licenses/W3C.html",
    "W3C-19980720": "https://spdx.org/licenses/W3C-19980720.html",
    "W3C-20150513": "https://spdx.org/licenses/W3C-20150513.html",
    "WTFPL": "https://spdx.org/licenses/WTFPL.html",
    "Watcom-1.0": "https://spdx.org/licenses/Watcom-1.0.html",
    "Wsuipa": "https://spdx.org/licenses/Wsuipa.html",
    "X11": "https://spdx.org/licenses/X11.html",
    "XFree86-1.1": "https://spdx.org/licenses/XFree86-1.1.html",
    "XSkat": "https://spdx.org/licenses/XSkat.html",
    "Xerox": "https://spdx.org/licenses/Xerox.html",
    "Xnet": "https://spdx.org/licenses/Xnet.html",
    "YPL-1.0": "https://spdx.org/licenses/YPL-1.0.html",
    "YPL-1.1": "https://spdx.org/licenses/YPL-1.1.html",
    "ZPL-1.1": "https://spdx.org/licenses/ZPL-1.1.html",
    "ZPL-2.0": "https://spdx.org/licenses/ZPL-2.0.html",
    "ZPL-2.1": "https://spdx.org/licenses/ZPL-2.1.html",
    "Zed": "https://spdx.org/licenses/Zed.html",
    "Zend-2.0": "https://spdx.org/licenses/Zend-2.0.html",
    "Zimbra-1.3": "https://spdx.org/licenses/Zimbra-1.3.html",
    "Zimbra-1.4": "https://spdx.org/licenses/Zimbra-1.4.html",
    "Zlib": "https://spdx.org/licenses/Zlib.html",
    "bzip2-1.0.5": "https://spdx.org/licenses/bzip2-1.0.5.html",
    "bzip2-1.0.6": "https://spdx.org/licenses/bzip2-1.0.6.html",
    "copyleft-next-0.3.0": "https://spdx.org/licenses/copyleft-next-0.3.0.html",
    "copyleft-next-0.3.1": "https://spdx.org/licenses/copyleft-next-0.3.1.html",
    "curl": "https://spdx.org/licenses/curl.html",
    "diffmark": "https://spdx.org/licenses/diffmark.html",
    "dvipdfm": "https://spdx.org/licenses/dvipdfm.html",
    "eCos-2.0": "https://spdx.org/licenses/eCos-2.0.html",
    "eGenix": "https://spdx.org/licenses/eGenix.html",
    "gSOAP-1.3b": "https://spdx.org/licenses/gSOAP-1.3b.html",
    "gnuplot": "https://spdx.org/licenses/gnuplot.html",
    "iMatix": "https://spdx.org/licenses/iMatix.html",
    "libpng-2.0": "https://spdx.org/licenses/libpng-2.0.html",
    "libtiff": "https://spdx.org/licenses/libtiff.html",
    "mpich2": "https://spdx.org/licenses/mpich2.html",
    "psfrag": "https://spdx.org/licenses/psfrag.html",
    "psutils": "https://spdx.org/licenses/psutils.html",
    "wxWindows": "https://spdx.org/licenses/wxWindows.html",
    "xinetd": "https://spdx.org/licenses/xinetd.html",
    "xpp": "https://spdx.org/licenses/xpp.html",
    "zlib-acknowledgement": "https://spdx.org/licenses/zlib-acknowledgement.html",
}

ALL_LICENSE_DICTS = [SPDX_LICENSES]

LICENSE_DICT = dict(
    itertools.chain.from_iterable(dct.items() for dct in ALL_LICENSE_DICTS)
)

__all__ = [LICENSE_DICT]



import json
from itertools import combinations, product

from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.parser import JSONParser_1_4
from omi.dialects.rdf.compiler import RDFCompiler
from omi.dialects.rdf.parser import RDFParser

from .test_compiler import assert_equal


def test_roundtrip():
    with open("tests/data/metadata_v14.json", "r") as _input_file:
        input_string = _input_file.read()
        assert_roundtrip(json.loads(input_string), input_string)


def assert_roundtrip(expected_json, input_string):
    json_compiler = JSONCompiler()
    json_parser = JSONParser_1_4()
    rdf_compiler = RDFCompiler()
    rdf_p = RDFParser()
    # Step 1: Parse JSON to internal structure
    internal_metadata = json_parser.parse_from_string(input_string)
    # Step 2: Translate to rdf
    _ = rdf_compiler.visit(internal_metadata)
    # Step 3: Parse rdf string
    internal_metadata2 = rdf_p.parse(rdf_compiler.graph)
    # Step 4: Translate to JSON
    result_json = json_compiler.visit(internal_metadata2)
    # Final step: Compare
    assert_equal(expected_json, result_json, disregard_ordering=True)


def __mask(jsn, keep):
    if isinstance(jsn, dict):
        return [dict(zip(keys, items))
                for keys in combinations(jsn.keys(),2)
                for items in product(*(__mask(jsn[field], keep)
                                       if field not in keep else [jsn[field]]
                                       for field in keys))
                if keys and all(k not in jsn.keys() or k in keys for k in keep) ]
    if isinstance(jsn, list):
        if jsn:
            return [[possibility] + jsn[1:] for possibility in __mask(jsn[0], keep)]
        else:
            return [[]]
    else:
        return [jsn]


def test_roundtrip_with_missing_fields():
    json_compiler = JSONCompiler()
    json_parser = JSONParser_1_4()
    with open("tests/data/metadata_v14.json", "r") as _input_file:
        input_string = _input_file.read()
        for expected_json in __mask(json.loads(input_string), keep=["id", "metaMetadata"]):
            assert_equal(json_compiler.visit(json_parser.parse(expected_json)), expected_json)




import json
from collections import OrderedDict

from omi import structure
from omi.oem_structures import oem_v15
from omi.dialects.base.compiler import Compiler


class JSONCompiler(Compiler):
    __METADATA_VERSION = "OEP-1.4.0"

    def _compile_date(self, date, format):
        if date:
            return date.strftime(format)
        else:
            return None

    def _construct_dict(self, *args, omit_none=True, **kwargs):
        """
        Accepts a list of arguments of shape (name: str, field: Compileable) and returns a dictionary that maps
        name -> self.visit(field). If `omit_none` is true, fields that are `None` are ignored.
        Parameters
        ----------
        args
        omit_none
        kwargs

        Returns
        -------

        """
        d = {
            field_name: self.visit(field)
            for field_name, field in args
            if (not omit_none) or (field is not None)
        }
        d.update(**kwargs)
        return d

    def visit_context(self, context: structure.Context, *args, **kwargs):
        result = self._construct_dict(
            ("homepage", context.homepage),
            ("documentation", context.documentation),
            ("sourceCode", context.source_code),
            ("contact", context.contact),
            ("grantNo", context.grant_number),
        )
        if context.funding_agency is not None:
            if context.funding_agency.name is not None:
                result["fundingAgency"] = context.funding_agency.name
            if context.funding_agency.logo is not None:
                result["fundingAgencyLogo"] = context.funding_agency.logo
        if context.publisher is not None:
            result["publisherLogo"] = context.publisher.logo
        return result

    def visit_contribution(self, contribution: structure.Contribution, *args, **kwargs):
        return self._construct_dict(
            ("title", contribution.contributor.name),
            ("email", contribution.contributor.email),
            ("object", contribution.object),
            ("comment", contribution.comment),
            ("date", self._compile_date(contribution.date, "%Y-%m-%d")),
        )

    def visit_language(self, language: structure.Language, *args, **kwargs):
        return str(language)

    def visit_spatial(self, spatial: structure.Spatial, *args, **kwargs):
        return self._construct_dict(
            ("location", spatial.location),
            ("extent", spatial.extent),
            ("resolution", spatial.resolution),
        )

    def visit_timestamp_orientation(
        self, tso: structure.TimestampOrientation, *args, **kwargs
    ):
        if tso == structure.TimestampOrientation.left:
            return "left"
        elif tso == structure.TimestampOrientation.middle:
            return "middle"
        elif tso == structure.TimestampOrientation.right:
            return "right"
        else:
            raise NotImplementedError

    def visit_temporal(self, temporal: structure.Temporal, *args, **kwargs):
        start = None
        end = None
        if temporal.ts_start is not None:
            start = self._compile_date(temporal.ts_start, "%Y-%m-%dT%H:%M%z")[:-2]
        if temporal.ts_end is not None:
            end = self._compile_date(temporal.ts_end, "%Y-%m-%dT%H:%M%z")[:-2]
        return self._construct_dict(
            ("referenceDate", self._compile_date(temporal.reference_date, "%Y-%m-%d")),
            timeseries=self._construct_dict(
                ("start", start),
                ("end", end),
                ("resolution", temporal.ts_resolution),
                ("alignment", temporal.ts_orientation),
                ("aggregationType", temporal.aggregation),
            ),
        )

    def visit_source(self, source: structure.Source, *args, **kwargs):
        return self._construct_dict(
            ("title", source.title),
            ("description", source.description),
            ("path", source.path),
            ("licenses", source.licenses),
        )

    def visit_license(self, lic: structure.License, *args, **kwargs):
        return self._construct_dict(
            ("name", lic.identifier),
            ("title", lic.name),
            ("path", lic.path),
        )

    def visit_terms_of_use(self, terms_of_use: structure.TermsOfUse):
        license_kwargs = (
            self.visit(terms_of_use.license) if terms_of_use.license else {}
        )
        return self._construct_dict(
            ("instruction", terms_of_use.instruction),
            ("attribution", terms_of_use.attribution),
            **license_kwargs
        )

    def visit_resource(self, resource: structure.Resource, *args, **kwargs):
        return self._construct_dict(
            ("profile", resource.profile),
            ("name", resource.name),
            ("path", resource.path),
            ("format", resource.format),
            ("encoding", resource.encoding),
            ("schema", resource.schema),
            ("dialect", resource.dialect),
        )

    def visit_field(self, field: structure.Field, *args, **kwargs):
        return self._construct_dict(
            ("name", field.name),
            ("description", field.description),
            ("type", field.type),
            ("unit", field.unit),
        )

    def visit_schema(self, schema: structure.Schema, *args, **kwargs):
        return self._construct_dict(
            ("primaryKey", schema.primary_key),
            ("foreignKeys", schema.foreign_keys),
            ("fields", schema.fields),
        )

    def visit_dialect(self, dialect: structure.Dialect, *args, **kwargs):
        return self._construct_dict(
            ("delimiter", dialect.delimiter),
            ("decimalSeparator", dialect.decimal_separator),
        )

    def visit_foreign_key(self, foreign_key: structure.ForeignKey, *args, **kwargs):
        if foreign_key.references:
            source_fields, target_fields, target_resources = zip(
                *map(self.visit, foreign_key.references)
            )
            target_resource = target_resources[0]

            return self._construct_dict(
                ("fields", source_fields),
                reference=self._construct_dict(
                    ("resource", target_resource),
                    ("fields", target_fields),
                ),
            )
        else:
            raise Exception("Missing reference in foreign key")

    def visit_reference(self, reference: structure.Reference, *args, **kwargs):
        return (
            reference.source.name,
            reference.target.name,
            reference.target.resource.name,
        )

    def visit_review(self, review: structure.Review, *args, **kwargs):
        return self._construct_dict(("path", review.path), badge=review.badge)

    def visit_meta_comment(self, comment: structure.MetaComment, *args, **kwargs):
        return self._construct_dict(
            ("metadata", comment.metadata_info),
            ("dates", comment.dates),
            ("units", comment.units),
            ("languages", comment.languages),
            ("licenses", comment.licenses),
            ("review", comment.review),
            ("null", comment.none),
        )

    def visit_metadata(self, metadata: structure.OEPMetadata, *args, **kwargs):
        publication_date = None
        if metadata.publication_date is not None:
            publication_date = self._compile_date(metadata.publication_date, "%Y-%m-%d")
        return self._construct_dict(
            ("name", metadata.name),
            ("title", metadata.title),
            ("id", metadata.identifier),
            ("description", metadata.description),
            ("keywords", metadata.keywords),
            ("publicationDate", publication_date),
            ("context", metadata.context),
            ("spatial", metadata.spatial),
            ("temporal", metadata.temporal),
            ("review", metadata.review),
            ("_comment", metadata.comment),
            ("language", metadata.languages),
            ("sources", metadata.sources),
            ("licenses", metadata.license),
            ("contributors", metadata.contributions),
            ("resources", metadata.resources),
            metaMetadata=self._construct_dict(
                ("metadataVersion", self.__METADATA_VERSION),
                metadataLicense=self._construct_dict(
                    name="CC0-1.0",
                    title="Creative Commons Zero v1.0 Universal",
                    path="https://creativecommons.org/publicdomain/zero/1.0/",
                ),
            ),
        )


class JSONCompilerOEM15(JSONCompiler):
    """
    Compiles OEMetadata version JSONCompilerOEM15.__METADATA_VERSION .

    Args:
        JSONCompiler (object):  Parent Class provides compiler methods for all
                                metadata fields that have not been changed in
                                the metadata structure.
    """

    __METADATA_VERSION = "OEP-1.5.1"

    def visit(self, obj, *args, **kwargs):
        """
        Calls the respective compiler for :class:`~omi.structure.Compilable` objects
        respective to :attr:`Compilable.__compiler_name__`


        Parameters
        ----------
        obj
            Object to compile

        Returns
        -------
            Metadata representation of `obj`

        """
        if isinstance(obj, list):
            return [self.visit(x) for x in obj if x is not None]
        if isinstance(obj, oem_v15.Compilable):
            meth = getattr(self, "visit_{name}".format(name=obj.__compiler_name__))
            return meth(obj, *args, **kwargs)
        else:
            return obj

    def visit_subject(self, subject: oem_v15.Subject, *args, **kwargs):
        return self._construct_dict(("name", subject.name), ("path", subject.path))

    def visit_timestamp_orientation(
        self, tso: oem_v15.TimestampOrientation, *args, **kwargs
    ):
        if tso == oem_v15.TimestampOrientation.left:
            return "left"
        elif tso == oem_v15.TimestampOrientation.middle:
            return "middle"
        elif tso == oem_v15.TimestampOrientation.right:
            return "right"
        else:
            raise NotImplementedError

    def visit_timeseries(self, timeseries: oem_v15.Timeseries, *args, **kwargs):
        start = None
        end = None
        if timeseries.ts_start is not None:
            start = self._compile_date(timeseries.ts_start, "%Y-%m-%dT%H:%M%z")[:-2]
        if timeseries.ts_end is not None:
            end = self._compile_date(timeseries.ts_end, "%Y-%m-%dT%H:%M%z")[:-2]
        return self._construct_dict(
            ("start", start),
            ("end", end),
            ("resolution", timeseries.ts_resolution),
            ("alignment", timeseries.ts_orientation),
            ("aggregationType", timeseries.aggregation),
        )

    def visit_temporal(self, temporal: oem_v15.Temporal, *args, **kwargs):
        return self._construct_dict(
            ("referenceDate", self._compile_date(temporal.reference_date, "%Y-%m-%d")),
            ("timeseries", temporal.timeseries_collection),
        )

    def visit_is_about(self, is_about: oem_v15.IsAbout, *args, **kwargs):
        return self._construct_dict(("name", is_about.name), ("path", is_about.path))

    def visit_value_reference(
        self, value_reference: oem_v15.ValueReference, *args, **kwargs
    ):
        return self._construct_dict(
            ("value", value_reference.value),
            ("name", value_reference.name),
            ("path", value_reference.path),
        )

    def visit_field(self, field: oem_v15.Field, *args, **kwargs):
        return self._construct_dict(
            ("name", field.name),
            ("description", field.description),
            ("type", field.type),
            ("is_about", field.is_about),
            ("value_reference", field.value_reference),
            ("unit", field.unit),
        )

    def visit_meta_comment(self, comment: oem_v15.MetaComment, *args, **kwargs):
        return self._construct_dict(
            ("metadata", comment.metadata_info),
            ("dates", comment.dates),
            ("units", comment.units),
            ("languages", comment.languages),
            ("licenses", comment.licenses),
            ("review", comment.review),
            ("null", comment.null),
            ("todo", comment.todo),
        )

    def visit_metadata(self, metadata: oem_v15.OEPMetadata, *args, **kwargs):
        publication_date = None
        if metadata.publication_date is not None:
            publication_date = self._compile_date(metadata.publication_date, "%Y-%m-%d")
        return self._construct_dict(
            ("name", metadata.name),
            ("title", metadata.title),
            ("id", metadata.identifier),
            ("description", metadata.description),
            ("subject", metadata.subject),
            ("keywords", metadata.keywords),
            ("publicationDate", publication_date),
            ("context", metadata.context),
            ("spatial", metadata.spatial),
            ("temporal", metadata.temporal),
            ("review", metadata.review),
            ("_comment", metadata.comment),
            ("language", metadata.languages),
            ("sources", metadata.sources),
            ("licenses", metadata.license),
            ("contributors", metadata.contributions),
            ("resources", metadata.resources),
            ("@id", metadata.databus_identifier),
            ("@context", metadata.databus_context),
            metaMetadata=self._construct_dict(
                ("metadataVersion", self.__METADATA_VERSION),
                metadataLicense=self._construct_dict(
                    name="CC0-1.0",
                    title="Creative Commons Zero v1.0 Universal",
                    path="https://creativecommons.org/publicdomain/zero/1.0/",
                ),
            ),
        )
